/*************************************************************************
** Definition of extended range date classe
**
** Created : 031103
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
**********************************************************************/

#include "extdate.h"
#include <assert.h>
#include <time.h>


/*****************************************************************************
  ExtDate class
 *****************************************************************************/

/*****************************************************************************
 *
 * Concepts :
 * a day is given by its rank after (or before if negative) the origin-day.
 * The origin-day is 01/01/0000. Its rank is 1.
 * 
 * ***************************************************************************/

int	ExtDate::month_length[] = {31, 28, 31, 30,  31,  30,  31,  31,  30,  31,  30,  31};
int	ExtDate::month_origin[] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};

QString ExtDate::my_shortMonthNames[12] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};
QString ExtDate::my_shortDayNames[7] = {
	"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"
};
QString ExtDate::my_longMonthNames[12] = {
	"January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"
};
QString ExtDate::my_longDayNames[7] = {
	"Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"
};

int	ExtDate::my_invalid_day = 0;	// RAF : mettre une valeur vraiment invalide

/*
 * y = year
 * m = month (Jan : 1, Feb : 2 ...
 * d = day in [1...31]
 */
int ExtDate::dayRank(int y, int m, int d)
{
	int	a_day_number;
	a_day_number = y*365 + (y+3)/4 - (y-1)/100 + (y-1)/400;
	a_day_number += dayOfYear(y, m, d);
	return a_day_number;
}

ExtDate::ExtDate( int y, int m, int d)
{
	jd = dayRank(y, m, d);
}

bool ExtDate::isValid() const
{
	return (jd != my_invalid_day);
}

int ExtDate::year() const
{
	int	a_year = (int)(jd/365.25);	// first approximation
	if (dayRank(a_year, 1, 1) > jd)
	{
		for (a_year-- ; dayRank(a_year, 1, 1) > jd ; a_year--);
	}

	if (dayRank(a_year+1, 1, 1) <= jd)
	{
		for ( ; dayRank(a_year+1, 1, 1) <= jd ; a_year++);
	}
	return a_year;
}

int ExtDate::month() const
{
	int	a_year = year();
	int	a_res = jd - dayRank(a_year, 1, 0);
	int	a_month;
	int	a_leap_year = leapYear(a_year) ? 1 : 0;
	if (month_origin[1] > a_res) return 1;
	if (month_origin[2] > a_res) return 2;
	for (a_month = 12 ; (month_origin[a_month-1]+a_leap_year) >= a_res ; a_month--);
	return a_month;
}

int ExtDate::day() const
{
	return jd - dayRank(year(), month(), 0);
}

int ExtDate::dayOfWeek() const
{
	int a_day = ((jd + 5) % 7);
	return (a_day == 0) ? 7 : a_day;
}

int ExtDate::dayOfYear() const
{
	return jd - dayRank(year(), 1, 0);
}

int ExtDate::daysInMonth() const
{
	int	a_month = month();
	return (a_month == 1) ? (leapYear(year()) ? 29 : 28) : month_length[a_month] ;
}

int ExtDate::daysInYear() const
{
	return (leapYear(year()) ? 366 : 365);
}

int ExtDate::weekNumber( int *yearNum) const
{
	// the year and week number are those of the next Sunday.
	ExtDate	a_date;
	a_date.set_jd(jd - dayOfWeek() + 7);
	int	a_year = a_date.year();
	*yearNum = a_year;
	return 1 + a_date.dayOfYear()/7;
}

#ifndef QT_NO_TEXTDATE
QString ExtDate::shortMonthName( int month ) {return my_shortMonthNames[month-1];};
QString ExtDate::shortDayName( int weekday ) {return my_shortDayNames[weekday-1];};
QString ExtDate::longMonthName( int month ) {return my_longMonthNames[month-1];};
QString ExtDate::longDayName( int weekday ) {return my_longDayNames[weekday-1];};
#endif //QT_NO_TEXTDATE
#ifndef QT_NO_TEXTSTRING
#if !defined(QT_NO_SPRINTF)
QString ExtDate::toStringSimpleArg( char code, int nb ) const
{
	QString	a_res;
	char	temp[10];
	switch (code)
	{
		case 'd' :
			{
				switch (nb)
				{
					case 1 :
						sprintf(temp, "%d", day());
						a_res = temp;
						break;
					case 2 :
						sprintf(temp, "%02d", day());
						a_res = temp;
						break;
					case 3 :
						a_res = shortDayName(dayOfWeek());
						break;
					case 4 :
						a_res = longDayName(dayOfWeek());
						break;
					default :
						a_res = "?d?";
						break;
				}
			}
			break;

		case 'M' :
			{
				int	m = month();
				switch (nb)
				{
					case 1 :
						sprintf(temp, "%d", m);
						a_res = temp;
						break;
					case 2 :
						sprintf(temp, "%02d", m);
						a_res = temp;
						break;
					case 3 :
						a_res = shortMonthName(m);
						break;
					case 4 :
						a_res = longMonthName(m);
						break;
					default :
						a_res = "?M?";
						break;
				}
			}
			break;

		case 'y' :
			{
				int	y = year();
				switch (nb)
				{
					case 2 :
						sprintf(temp, "%02d", y%100);
						a_res = temp;
						break;
					case 4 :
						sprintf(temp, "%04d", y);
						a_res = temp;
						break;
					default :
						a_res = "?y?";
						break;
				}
			}
			break;

		default :
			break;
	}
	return a_res;
}

QString ExtDate::toString( Qt::DateFormat f) const
{
	QString	a_format;
	switch (f)
	{
		case Qt::TextDate :	// Sat May 20 1995
			a_format = "ddd MMM d yyyy";
			break;

		case Qt::ISODate :	// YYYY-MM-DD
			a_format = "yyyy-MM-dd";
			break;

		case Qt::LocalDate :	// local settings
			a_format = "toString : Qt::LocalDate not implemented yet";
			break;

		default :
			a_format = "toString : unknown format";
			break;

	}
	return toString(a_format);
}
#endif
QString ExtDate::toString( const QString& format ) const
{
	// d{1-4}, M{1-4}, yy|yyyy
	int	i, a_nb = 0;
	QString	a_res;
	char	a_code = '\0';
	char	a_char;
	for (i = 0 ; i <= format.length() ; i++)
	{
		if (i < format.length())
		{
			a_char = format[i];
		}
		else
		{
			a_char = '\0';
		}
		if (a_code == a_char)
		{
			a_nb++;
		}
		else
		{
			if ((a_code == 'd' || a_code == 'M' || a_code == 'y') && a_nb != 0)
			{
				a_res += toStringSimpleArg(a_code, a_nb);
			}
			if (a_char == 'd' || a_char == 'M' || a_char == 'y')
			{
				a_code = a_char;
				a_nb = 1;
			}
			else
			{
				a_code = '\0';
				a_nb = 0;
				if (a_char != '\0')
					a_res += a_char;
			}
		}
	}
	return a_res;
}
#endif
bool   ExtDate::setYMD( int y, int m, int d )
{
	jd = dayRank(y, m, d);
	return true;
}

ExtDate  ExtDate::addDays( int days ) const
{
	ExtDate	a_date;
	a_date.set_jd(jd + days);
	return a_date;
}

ExtDate  ExtDate::addMonths( int months ) const
{
	int	a_month = month() + months;
	int	a_year = year() + (a_month - 1)/12;	// month : [1..12]
	a_month  = 1 + (a_month -1) % 12;
	return ExtDate(a_year, a_month, day());
}

ExtDate  ExtDate::addYears( int years ) const
{
	return ExtDate(year() + years, month(), day());
}

int   ExtDate::daysTo( const ExtDate & a_date) const
{
	return a_date.jd - jd;	// RAF
}

ExtDate ExtDate::currentDate(Qt::TimeSpec ts)
{
	time_t		a_current_time;
	struct	tm	a_current_time_tm;

	time(&a_current_time);
	switch (ts)
	{
		case Qt::LocalTime :
			localtime_r(&a_current_time, &a_current_time_tm);
			break;

		case Qt::UTC :
			gmtime_r(&a_current_time, &a_current_time_tm);
			break;

		default :
			assert(0);
			break;
	}
	return ExtDate(a_current_time_tm.tm_year + 1900, a_current_time_tm.tm_mon + 1, a_current_time_tm.tm_mday);
}

#ifndef QT_NO_DATESTRING
ExtDate ExtDate::fromString( const QString& s, Qt::DateFormat f )
{
	return ExtDate(0, 0,0);	// RAF
}

#endif
bool	ExtDate::isValid( int y, int m, int d )
{
	if (m < 1 || m > 12) return false;
	if (d < 1) return false;
	if (m != 2 && d > month_length[m-1]) return false;
	if (d > (month_length[1] + (leapYear(y) ? 1 : 0))) return false;
	return true;
}

bool	ExtDate::leapYear( int year )
{
	// year is the year-number where JC birth is 0
	if ((year % 4) != 0) return false;
	// multiple of 4 : can be a leap year
	// centennial years are NOT leap, but quadri-centennial ARE.
	if ((year % 400) == 0) return true;
	if ((year % 100) == 0) return false;
	// year is multiple of 4 but not centennial so leap year !
	return true;
}

uint	ExtDate::gregorianToJulian( int y, int m, int d )
{
	return 0;	// RAF
}

void	ExtDate::julianToGregorian( uint jd, int &y, int &m, int &d )
{
}

int ExtDate::dayOfYear(int y, int m, int d)
{
	return month_origin[m-1] + d + ((m > 1) ? (leapYear(y) ? 1 : 0) : 0);
}

#ifndef QT_NO_DATASTREAM
Q_EXPORT QDataStream &operator<<( QDataStream & ostream, const ExtDate & date)
{
	return ostream;
}

Q_EXPORT QDataStream &operator>>( QDataStream & ostream, ExtDate & date)
{
	return ostream;
}

#endif // QT_NO_DATASTREAM
